
import os
import sys
import imp
import yaml
from .constants import *
from .logger import Logger
from .dict_utils import *
from pathlib import Path
from .exceptions import DSOException


_default_config = {
    'kind': 'dso/application/v1',
    'version': 1,
    'project': 'default',
    'application': 'default',
    'parameter': {
        'provider': {
            'id': '',
            'spec': {}
        },
    },
    'secret': {
        'provider': {
            'id': '',
            'spec': {}
        },
    },
    'template': {
        'provider': {
            'id': '',
            'spec': {}
        },
        'render': {}
    }
}


def get_default_config():
    return _default_config.copy()


class ConfigManager:
    @property
    def config_dir(self):
        return '.dso'

    @property
    def config_file(self):
        return 'dso.yml'

    @property
    def install_path(self):
        return os.path.dirname(os.path.abspath(__file__))
    
    working_dir = ''
    local_config = {}
    local_config_file_path = ''
    local_config_dir_path = ''
    global_config = {}
    global_config_file_path = ''
    global_config_dir_path = ''
    overriden_config = {}
    inherited_config = {}
    inherited_config_files = []
    merged_config = {}

    # def __init__(self):
    #     # path = os.path.join(os.path.expanduser("~"), self.config_dir, self.config_file)
    #     # if os.path.exists(path):
    #     #     self.__config = yaml.safe_load(open(path))

    def load_global_config(self):
        self.global_config_dir_path = os.path.join(Path.home(), self.config_dir)
        self.global_config_file_path = os.path.join(self.global_config_dir_path, self.config_file)

        if not os.path.exists(self.global_config_file_path):
            self.global_config = {}
            Logger.debug("Global DSO configuration not found.")
            return

        Logger.debug(f"Global DSO configuration found: path={self.global_config_file_path}")

        try:
            self.global_config = yaml.load(open(self.global_config_file_path, 'r', encoding='utf-8'), yaml.SafeLoader)
        except:
            raise DSOException(CLI_MESSAGES['InvalidDSOConfigurationFile'])

        self.update_merged_config()


    def load_local_config(self, show_warnings=True):
        self.local_config_dir_path = os.path.join(self.working_dir, self.config_dir)
        self.local_config_file_path = os.path.join(self.local_config_dir_path, self.config_file)

        if not os.path.exists(self.local_config_file_path):
            self.local_config = {}
            if show_warnings:
                Logger.warn(CLI_MESSAGES['NoDSOConfigFound'])
            return

        Logger.debug(f"Local DSO configuration found: path={self.local_config_file_path}")

        try:
            self.local_config = yaml.load(open(self.local_config_file_path, 'r', encoding='utf-8'), yaml.SafeLoader)
        except:
            raise DSOException(CLI_MESSAGES['InvalidDSOConfigurationFile'])

        self.update_merged_config()

    def load_inherited_config(self):

        for dir in Path(self.working_dir).resolve().parents:
            configFilePath = os.path.join(dir, self.config_dir, self.config_file)
            if os.path.exists(configFilePath):
                if not os.path.samefile(configFilePath, self.global_config_file_path):
                    Logger.debug(f"Inherited DSO configuration found: path={configFilePath}")
                    self.inherited_config_files.append(configFilePath)

        self.inherited_config = {}

        for configFile in reversed(self.inherited_config_files):
            try:
                config = yaml.load(open(configFile, 'r', encoding='utf-8'), yaml.SafeLoader)
            except:
                raise DSOException(CLI_MESSAGES['InvalidDSOConfigurationFile'])
            
            self.inherited_config = merge_dicts(config, self.inherited_config)

        self.update_merged_config()

    def override_config(self, config_overrides):

        self.overriden_config = {}

        if config_overrides:
            try:
                configs = config_overrides.split(',')
                for config in configs:
                    key = config.split('=')[0].strip()
                    value = config.split('=')[1].strip()
                    old = get_dict_item(self.merged_config, key.split('.'))
                    if not old:
                        Logger.warn(f"DSO configuration '{key}' was not found in the merged configuration.")
                    else:
                        Logger.warn(f"DSO configuration '{key}' was overriden from '{old}' to '{value or 'default'}'.")
                    set_dict_value(self.overriden_config, key.split('.'), value)
            except:
                raise DSOException(CLI_MESSAGES['InvalidDSOConfigOverrides'])

        self.update_merged_config()


    def load(self, working_dir, config_overrides):
        self.working_dir = working_dir
        self.load_local_config()
        self.load_global_config()
        self.load_inherited_config()
        self.override_config(config_overrides)
        self.check_version()

    def check_version(self):
        if not int(self.merged_config['version']) == int(_default_config['version']):
            if int(self.merged_config['version']) > int(_default_config['version']):
                Logger.warn(CLI_MESSAGES['DSOConfigNewer'].format(_default_config['version'], self.merged_config['version']))
            else:
                Logger.warn(CLI_MESSAGES['DSOConfigOlder'].format(_default_config['version'], self.merged_config['version']))


    def get_provider_default_spec(self, provider_id):
        if provider_id in sys.modules:
            provider = sys.modules[provider_id]
            # provider = __import__(provider_id)
            # provider = imp.find_module(provider_id)
            # Logger.debug(f"Providers '{provider_id}' has already been loaded.")
        else:
            Logger.debug(f"Providers '{provider_id}' has not been loaded. Loading...")
            providerPackagePath = os.path.join(self.install_path, 'provider', provider_id)
            if not os.path.exists(providerPackagePath):
                raise DSOException(f"Providers '{provider_id}' not found.")
            provider = imp.load_package(provider_id, providerPackagePath)
        
        return provider.get_default_spec()


    def update_merged_config(self):
        self.merged_config = get_default_config()
        self.merged_config = merge_dicts(self.global_config, self.merged_config)
        self.merged_config = merge_dicts(self.inherited_config, self.merged_config)
        self.merged_config = merge_dicts(self.local_config, self.merged_config)
        self.merged_config = merge_dicts(self.overriden_config, self.merged_config)
        ### add missing default specs for each provider
        providers = ['parameter', 'secret', 'template']
        for provider in providers:
            providerId = self.merged_config[provider]['provider']['id']
            if providerId:
                save = self.merged_config[provider]['provider']['spec'].copy()
                self.merged_config[provider]['provider']['spec'] = self.get_provider_default_spec(f"{provider}/{providerId}")
                merge_dicts(save, self.merged_config[provider]['provider']['spec'])

    def flush_local_config(self):
        os.makedirs(self.local_config_dir_path, exist_ok=True)
        with open(self.local_config_file_path, 'w') as outfile:
            yaml.dump(self.local_config, outfile, sort_keys=False, indent=2)

    def flush_global_config(self):
        os.makedirs(self.global_config_dir_path, exist_ok=True)
        with open(self.global_config_file_path, 'w') as outfile:
            yaml.dump(self.global_config, outfile, sort_keys=False, indent=2)

    @property
    def application(self):
        if 'application' in self.merged_config:
            result = self.merged_config['application'] or 'default'
        else:
            result  = os.getenv('DSO_APPLICATION') or 'default'
        if self.project == 'default':
            if not result == 'default':
                Logger.warn(f"The specific context of application '{result}' was ignored because the global context was used.")
                result = 'default'
        return result.lower()

    @property
    def project(self):
        if 'project' in self.merged_config:
            result = self.merged_config['project'] or 'default'
        else:
            result  = os.getenv('DSO_PROJECT') or 'default'
        return result.lower()

    @property
    def parameter_provider(self):
        try:
            result = self.merged_config['parameter']['provider']['id']
        except KeyError:
            raise DSOException(CLI_MESSAGES['ProviderNotSet'].format('Parameter'))
        return result

    @property
    def secret_provider(self):
        try:
            result = self.merged_config['secret']['provider']['id']
        except KeyError:
            raise DSOException(CLI_MESSAGES['ProviderNotSet'].format('Secret'))
        return result

    @property
    def template_provider(self):
        try:
            result = self.merged_config['template']['provider']['id']
        except KeyError:
            raise DSOException(CLI_MESSAGES['ProviderNotSet'].format('Template'))
        return result

    def parameter_spec(self, key=None):
        try:
            result = self.merged_config['parameter']['provider']['spec']
        except KeyError:
            raise DSOException(CLI_MESSAGES['ProviderNotSet'].format('Parameter'))
        if not key:
            return result
        elif key in result:
            return result[key]
        else:
            return None

    def secret_spec(self, key=None):
        try:
            result = self.merged_config['secret']['provider']['spec']
        except KeyError:
            raise DSOException(CLI_MESSAGES['ProviderNotSet'].format('Secret'))
        if not key:
            return result
        elif key in result:
            return result[key]
        else:
            return None

    def template_spec(self, key):
        try:
            result = self.merged_config['template']['provider']['spec']
        except KeyError:
            raise DSOException(CLI_MESSAGES['ProviderNotSet'].format('Template'))
        if not key:
            return result
        elif key in result:
            return result[key]
        else:
            return None

    def get_template_render_paths(self, key=None):
        result = get_dict_item(self.local_config, ['template', 'render']) or {}
        if not key:
            return result
        else:
            return {x:result[x] for x in result if x==key}

    def register_template_custom_render_path(self, key, render_path):
        result = get_dict_item(self.local_config, ['template', 'render']) or {}
        # if os.path.isabs(render_path):
        #     raise DSOException(CLI_MESSAGES['AbsTemplateRenderPath'].format(render_path))
        if os.path.isdir(render_path):
            raise DSOException(CLI_MESSAGES['InvalidRenderPathExistingDir'].format(render_path))
        result[key] = render_path
        self.local_config['template']['render'] = result
        self.flush_local_config()

    def unregister_template_custom_render_path(self, key):
        result = get_dict_item(self.local_config, ['template', 'render'])
        if key in result:
            self.local_config['template']['render'].pop(key)
            self.flush_local_config()

    def get(self, key=None, scope=''):
        if key:
            Logger.info("Getting '{0}' from DSO configurations...".format(key))
        else:
            Logger.info("Getting DSO configurations...")

        if scope == 'local':
            usedConfig = merge_dicts(self.overriden_config, self.local_config.copy())
        elif scope == 'global':
            usedConfig = merge_dicts(self.overriden_config, self.global_config.copy())
        else:
            usedConfig = self.merged_config.copy()

        if key:
            result = get_dict_item(usedConfig, key.split('.'))
            if not result:
                raise DSOException(f"'{key}' has not been set in the DSO configurations.")
            return result
        else:
            return usedConfig

    def set(self, key, value, use_global):
        if use_global:
            Logger.info(f"Setting '{key}' to '{value}' in the global DSO configurations...")
            if not os.path.exists(self.local_config_file_path):
                raise DSOException("The global configuration has not been intitialized. Run 'dso config init --global' to initialize it.")

            set_dict_value(self.global_config, key.split('.'), value, overwrite_parent=True, overwrite_children=True)
            self.update_merged_config()
            self.flush_global_config()
        else:
            Logger.info(f"Setting '{key}' to '{value}' in the local DSO configurations...")
            if not os.path.exists(self.local_config_file_path):
                raise DSOException("The working directory has not been intitialized. Run 'dso config init' to initialize it.")

            set_dict_value(self.local_config, key.split('.'), value, overwrite_parent=True, overwrite_children=True)
            self.update_merged_config()
            self.flush_local_config()

    def delete(self, key, use_global):
        if use_global:
            Logger.info(f"Deleting '{key}' from the global DSO configurations...")
            parent = get_dict_item(self.global_config, key.split('.')[:-1])
            if parent and key.split('.')[-1] in parent:
                parent.pop(key.split('.')[-1])
                self.flush_global_config()
                self.update_merged_config()
            else:
                raise DSOException(f"'{key}' not found in the global DSO configuratoins.")

        else:
            Logger.info(f"Deleting '{key}' from the local DSO configurations...")
            parent = get_dict_item(self.local_config, key.split('.')[:-1])
            if parent and key.split('.')[-1] in parent:
                parent.pop(key.split('.')[-1])
                self.flush_local_config()
                self.update_merged_config()
            else:
                raise DSOException(f"'{key}' not found in the local DSO configuratoins.")

    def init(self, working_dir, init_config, config_overrides, use_local):
        Logger.info("Initializing DSO configurations...")
        self.working_dir = working_dir

        self.local_config_dir_path = os.path.join(self.working_dir, self.config_dir)
        self.local_config_file_path = os.path.join(self.local_config_dir_path, self.config_file)

        if os.path.exists(self.local_config_file_path):
            Logger.warn("The working directory has already been initialized.")

        config = {}
        
        ### use init_config instead of local/inherited config
        if init_config:
            config = merge_dicts(init_config, config)
            ### merge with existing local configuration?
            if use_local:
                Logger.debug("Merging local configuration...")
                self.load_local_config(False)
                config = merge_dicts(self.local_config, config)
        else:
            ### do not show warning if directory is not initialized yet
            self.load_local_config(False)
            config = merge_dicts(self.local_config, config)
            ### override locally inherited configuration?
            if use_local: 
                Logger.debug("Merging global configuration...")
                self.load_global_config()
                config = merge_dicts(self.global_config, config)

                Logger.debug("Merging inherited configuration...")
                self.load_inherited_config()
                config = merge_dicts(self.inherited_config, config)

        ### if config overrides, merge them to local
        if config_overrides:
            Logger.debug("Merging configuration overrides...")
            self.override_config(config_overrides)
            config = merge_dicts(self.overriden_config, config)

        self.local_config = config
        self.update_merged_config()

        self.flush_local_config()



Configs = ConfigManager()





from .logger import Logger
from .exceptions import DSOException

def clean_directory(path):
    for path in Path(path).glob("**/*"):
        if path.is_file():
            path.unlink()
        elif path.is_dir():
            rmtree(path)


###--------------------------------------------------------------------------------------------

def merge_dicts(source, destination):
    if not source: return destination
    for key, value in source.items():
        if isinstance(value, dict):
            if key in destination.keys():
                if not isinstance(destination[key], dict):
                    raise DSOException(f"Faile to merge '{key}' beacuse destination has an existing key with incompatible type ({type(destination[key])}) to that of the source ({type(source[key])}).")
            else:
                destination[key] = {}
            node = destination[key]
            merge_dicts(value, node)
        else:
            destination[key] = value

    return destination


###--------------------------------------------------------------------------------------------

def flatten_dict(input_node: dict, prefixed_key = '', delimiter = '.', output_dict: dict = {}):
    if isinstance(input_node, dict):
        for key, val in input_node.items():
            new_key = f"{prefixed_key}{delimiter}{key}" if prefixed_key else f"{key}"
            flatten_dict(val, new_key, delimiter, output_dict)
    elif isinstance(input_node, list):
        for idx, item in enumerate(input_node):
            flatten_dict(item, f"{prefixed_key}{delimiter}{idx}", delimiter, output_dict)
    else:
        output_dict[prefixed_key] = input_node
    return output_dict

###--------------------------------------------------------------------------------------------

def deflatten_dict(input: dict, delimiter = '.'):
    data = {}
    for key, value in input.items():
        set_dict_value(data, key.split(delimiter), value, overwrite_parent=True, overwrite_children=True)
    return data

###--------------------------------------------------------------------------------------------

def get_dict_item(dic, keys, create=True):
    for i in range(0, len(keys)):
        key = keys[i]
        if isinstance(dic, dict):
            if not key in dic.keys():
                if create:
                    dic[key] = {}
                else:
                    return None
            dic = dic[key]
        elif isinstance(dic, list):
            raise DSOException("Lists items are not allowed '{0}'. Must be converted to dictionary.".format('.'.join(keys[0:i])))
        else:
            return None
    return dic
    

###--------------------------------------------------------------------------------------------

def set_dict_value(dic, keys, value, overwrite_parent=False, overwrite_children=False):
    parent_item = get_dict_item(dic, keys[:-1])
    lastKey = keys[-1]
    ### parent item is expected to be a dictionary
    if not isinstance(parent_item, dict):
        if overwrite_parent:
            grand_parent_item = get_dict_item(dic, keys[:-2])
            grand_parent_item[keys[-2]] = {}
            parent_item = grand_parent_item[keys[-2]]
            Logger.warn("'{0}' was overwritten by '{1}.".format('.'.join(keys[:-1]),'.'.join(keys)))
        else:
            raise DSOException("Failed to set '{0}' becasue it is a '{1}'. Dictionary type was expected.".format('.'.join(keys), type(parent_item)))
    if lastKey in parent_item.keys():
        ### item is expected to be a basic type (string, number, ...)
        if isinstance(parent_item[lastKey], dict) or isinstance(parent_item[lastKey], list) or isinstance(parent_item[lastKey], set) or isinstance(parent_item[lastKey], tuple):
            if overwrite_children:
                Logger.warn("'{0}' was overwritten.".format('.'.join(keys)))
            else:
                raise DSOException("Failed to set '{0}' becasue it is a '{1}'. Simple type was expected.".format('.'.join(keys), type(parent_item)))
    parent_item[lastKey] = value

###--------------------------------------------------------------------------------------------

def del_dict_item(dic, keys):
    item = get_dict_item(dic, keys[:-1], create=False)
    if not (item and keys[-1] in item.keys()):
        return False

    if isinstance(item[keys[-1]], dict):
        raise DSOException("'{0}' is a non-empty scope and cannot be deleted.".format('.'.join(keys)))

    item.pop(keys[-1])
    return True

###--------------------------------------------------------------------------------------------

def del_dict_empty_item(dic, keys):
    if not (dic and len(dic.keys()) > 0): return
    item = get_dict_item(dic, keys)
    if len(item) == 0:
        item = get_dict_item(dic, keys[:-1])
        item.pop(keys[-1])
        del_dict_empty_item(dic, keys[:-1])


###--------------------------------------------------------------------------------------------

def is_binary_file(filename):
    """ 
    Return true if the given filename appears to be binary.
    File is considered to be binary if it contains a NULL byte.
    FIXME: This approach incorrectly reports UTF-16 as binary.
    """
    with open(filename, 'rb') as f:
        for block in f:
            if b'\0' in block:
                return True
    return False



from .cli import cli
from .version import __version__
from .exceptions import *
from .config import Configs
from .providers import ProviderBase, StoreProvider, ProviderManager, Providers
from .parameters import ParameterProvider, ParameterManager, Parameters
from .secrets import SecretProvider, SecretManager, Secrets
from .templates import TemplateProvider, TemplateManager, Templates




from .logger import Logger
from .config import Configs
from .providers import StoreProvider, Providers
from .stages import Stages

class ParameterProvider(StoreProvider):
    def list(self, stage, uninherited, format):
        raise NotImplementedError()
    def add(self, stage, key, value):
        raise NotImplementedError()
    def delete(self, stage, key):
        raise NotImplementedError()
    def get(self, stage, key):
        raise NotImplementedError()

class ParameterManager():

    def validate_key(self, key):
        provider = Providers.ParameterProvider()
        Logger.info("Start validating parameter key...")
        return provider.validate_key(key)

    def list(self, stage, uninherited):
        project = Configs.project
        application = Configs.application
        provider = Providers.ParameterProvider()
        if uninherited:
            Logger.info(f"Start listing uninherited SSM parameters: project={project}, application={application}, stage={Stages.shorten(stage)}")
        else:
            Logger.info(f"Start listing SSM parameters: project={project}, application={application}, stage={Stages.shorten(stage)}")
        return provider.list(project, application, stage, uninherited)

    def add(self, stage, key, value):
        self.validate_key(key)
        project = Configs.project
        application = Configs.application
        provider = Providers.ParameterProvider()
        Logger.info(f"Start adding parameter: project={project}, application={application}, stage={Stages.shorten(stage)}, key={key}, value={value}")
        return provider.add(project, application, stage, key, value)

    def get(self, stage, key, revision):
        # self.validate_key(key)
        project = Configs.project
        application = Configs.application
        provider = Providers.ParameterProvider()
        Logger.info(f"Start getting parameter value: project={project}, application={application}, stage={Stages.shorten(stage)}, key={key}")
        return provider.get(project, application, stage, key, revision)

    def history(self, stage, key):
        # self.validate_key(key)
        project = Configs.project
        application = Configs.application
        provider = Providers.ParameterProvider()
        Logger.info(f"Start getting parameter history: project={project}, application={application}, stage={Stages.shorten(stage)}, key={key}")
        return provider.history(project, application, stage, key)

    def delete(self, stage, key):
        # self.validate_key(key)
        project = Configs.project
        application = Configs.application
        provider = Providers.ParameterProvider()
        Logger.info(f"Start deleting parameter: project={project}, application={application}, stage={Stages.shorten(stage)}, key={key}")
        return provider.delete(project, application, stage, key)

Parameters = ParameterManager()


#!/usr/bin/self.env python
import io
import os
import sys
from shutil import rmtree
from setuptools import find_packages, setup, Command
import requests
import json

# Package meta-data.
NAME = 'dsocli'
DESCRIPTION = 'DSO CLI'
REQUIRES_PYTHON = '>=3.0'
URL = "https://github.com/ramtinkazemi/dsocli"
EMAIL = 'ramtin.kazemi@gmail.com'
AUTHOR = 'Ramtin Kazemi'
MAJOR_VERSION = 1
MINOR_VERSION = 0

# ------------------------------------------------------------------------
# ------------------------------------------------------------------------
# ------------------------------------------------------------------------

here = os.path.abspath(os.path.dirname(__file__))

# ------------------------------------------------------------------------

### set_long_description
try:
    with io.open(os.path.join(here, 'PyPI-Description.md'), encoding='utf-8') as f:
        long_description = '\n' + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

### set_pypi_index
git_branch = os.popen('git symbolic-ref --short HEAD').read().strip()
if git_branch == 'master':
    env='prod'
    pypi_index='pypi'
    pypi_url='https://pypi.org/pypi/dsocli/json'
elif git_branch == 'develop':
    env='dev'
    pypi_index='testpypi'
    pypi_url='https://test.pypi.org/pypi/dsocli/json'
else:
    status("Only 'develop' and 'master' git_branches can be deployed.")
    sys.exit(1)

### set version
published_versions = sorted(list(requests.get(pypi_url).json()['releases'].keys()), reverse=True)
if not published_versions: 
    current_version = f"{MAJOR_VERSION}.{MINOR_VERSION}.0"
else:
    current_version = published_versions[0]

current_major = int(current_version.split('.')[0])
current_minor = int(current_version.split('.')[1])
current_release = int(current_version.split('.')[2])
new_major = MAJOR_VERSION
new_minor = MINOR_VERSION
if new_major < current_major:
    raise Exception(f"New version '{new_major}.{new_minor}.x' cannot be smaller then the current version '{current_major}.{current_minor}.x'.")
elif new_major == current_major:
    if new_minor < current_minor:
        raise Exception(f"New version '{new_major}.{new_minor}.x' cannot be smaller then the current version '{current_major}.{current_minor}.x'.")
    if new_minor == current_minor:
        new_release = int(current_release) + 1
    else:
        new_release = 1
else:
    new_release = 1

new_version = f"{new_major}.{new_minor}.{new_release}"

with open(os.path.join(here, 'src', NAME, 'package.json'), 'r') as f:
    package_info = json.load(f)
package_info['version'] = new_version
with open(os.path.join(here, 'src', NAME, 'package.json'), 'w') as f:
    json.dump(package_info, f, sort_keys=False, indent=2)

# ------------------------------------------------------------------------

class PublishCommand(Command):
    """Support setup.py publish."""

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass
        
    def run(self):
        
        self.status(f"Git Branch={git_branch}\nEnv={env}\nPyPI Index={pypi_index}\nCurrent Version={current_version}\nNew Version={new_version}")
        try:
            self.status('Removing previous builds…')
            rmtree(os.path.join(here, 'dist'))
            rmtree(os.path.join(here, 'build'))
        except OSError:
            pass

        self.status('Building Source and Wheel (universal) distribution…')
        os.system('{0} setup.py sdist bdist_wheel --universal'.format(sys.executable))

        self.status('Uploading the package to PyPI via Twine…')
        ### read from ~/.pypirc
        # TWINE_PASSWORD=$(<pypi-token.secret) python -m twine upload --username __token__ --skip-existing --repository testpypi dist/*
        os.system('{0} -m twine upload --skip-existing --repository {1} dist/*'.format(sys.executable, pypi_index))

        # self.status('Pushing git tags…')
        # os.system('git tag v{0}'.format(new_version))
        # os.system('git push --tags')

        sys.exit()

# ------------------------------------------------------------------------
# ------------------------------------------------------------------------
# ------------------------------------------------------------------------


setup(
    name=NAME,
    version=new_version,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    author_email=EMAIL,
    url=URL,
    project_urls={
        'Documentation': URL,
        'Source': URL
    },
    packages=find_packages('src', exclude=["tests", "*.tests", "*.tests.*", "tests.*"]),
    package_dir = { '' : 'src/' }, 
    setup_requires=['wheel'],
    python_requires=REQUIRES_PYTHON,
    install_requires=open(os.path.join(here, 'requirements/prod.in'), 'r').readlines(),
    extras_require=None,
    package_data={'': ['package.json']},
    include_package_data=True,
    license='GPLV3',
    license_files = 'LICENSE.md',
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy'
    ],
    entry_points={
        # 'distutils.commands': ['publish = package.module:publish_command'],
        'console_scripts': [f'dso = {NAME}.cli:cli'],
        },
    cmdclass={'publish': PublishCommand},
)

print('Installing the new version...')
os.system('pip iuninstall dsocli -y')
os.system('pip install dsocli')


import os
import re
import yaml
import json
import pathlib
import numbers
from dsocli.logger import Logger
from dsocli.config import Configs
from dsocli.parameters import ParameterProvider
from dsocli.stages import Stages
from dsocli.constants import *
from dsocli.utils import *

defaults = {
    'format': 'yaml',
    'allowGroups': 'no',
    'prependGroups': 'yes',
    'groupDelimiter': '/',
    'nestedDelimiter': '.',
}

config = {
    'parameters_dir' : '{0}/parameters'.format(Configs.dir_name),
    'file_extension' : {
        'yaml' : 'yml',
        'json' : 'json',
    }
}

class LocalParameterProvider(ParameterProvider):

    def __init__(self):
        super().__init__('parameter/local/v1')

###--------------------------------------------------------------------------------------------
###--------------------------------------------------------------------------------------------

    def __get_param_file_extension(self):
        format = Configs.parameter_spec('format')
        if format is None:
            Logger.debug(f"'format' is not set for the parameter provider, defaulted to '{defaults['format']}'.")
            format = defaults['format']

        return config['file_extension'][format]

###--------------------------------------------------------------------------------------------

    def __get_params_root_path(self, context=None):
        if not context:
            return f"{Configs.working_dir}/{config['parameters_dir']}"
        stage, env = Stages.parse_context(context)
        # if env == 'default':
        #     return f"{params_root_path}/{stage}"
        # else:
        #     return f"{params_root_path}/{stage}/{env}"
        return f"{Configs.working_dir}/{config['parameters_dir']}/{stage}/{env}"

###--------------------------------------------------------------------------------------------

    def __parse_parameter_key(self, key):
        if '/' in key:
            keySplitted = key.split('/')
            name = keySplitted[-1]
            group = '/'.join(keySplitted[:-1])
        else:
            group = 'default_group'
            name = key

        return group, name

###--------------------------------------------------------------------------------------------

    def __get_param_group_from_key(self, key):
        return self.__parse_parameter_key(key)[0]

###--------------------------------------------------------------------------------------------

    def __get_param_name_from_key(self, key):
        return self.__parse_parameter_key(key)[1]

###--------------------------------------------------------------------------------------------

    def __parse_group_from_file_name(self, filename):
        return os.path.splitext(filename)[0]  ### get rid of extension

###--------------------------------------------------------------------------------------------

    def __get_param_file_name_from_key(self, key):
        fileName = self.__get_param_group_from_key(key)
        fileExt = self.__get_param_file_extension()
        return f"{fileName}.{fileExt}"


###--------------------------------------------------------------------------------------------

    def __get_param_file_path_from_key(self, context, key, create=False):
        root_path = self.__get_params_root_path(context)
        fileName = self.__get_param_file_name_from_key(key)
        filePath = f"{root_path}/{fileName}"
        if not os.path.exists(filePath):
            if create:
                pathlib.Path(os.path.dirname(filePath)).mkdir(parents=True, exist_ok=True) 
                open(filePath, 'w').close()
            else:
                return None
        return filePath

###--------------------------------------------------------------------------------------------

    def __load_param_file(self, path):
        fileExt = pathlib.Path(path).suffix
        if fileExt in ['.yml','.yaml']:
            return yaml.safe_load(open(path)) or {}
        elif fileExt in ['.json']:
            return json.load(open(path)) or {}

###--------------------------------------------------------------------------------------------

    def __save_param_file(self, data, path):
        fileExt = pathlib.Path(path).suffix
        if fileExt in ['.yml','.yaml']:
            with open(path, 'w') as outfile:
                yaml.dump(data, outfile, sort_keys=False, indent=2)
        else:
            with open(path, "w") as outfile: 
                json.dump(data, outfile, sort_keys=False, indent=2)

###--------------------------------------------------------------------------------------------

    def __list_param_file_names(self, path):
        result = []
        if os.path.exists(path):
            fileExt = self.__get_param_file_extension()
            for fileName in os.listdir(path):
                if fileName.endswith(fileExt):
                    result.append(fileName)        
        return result

###--------------------------------------------------------------------------------------------
 
    def __context_exists(self, context):
       return os.path.exists(self.__get_params_root_path(context))

###--------------------------------------------------------------------------------------------

    def __param_group_exists(self, context, key):
        return not self.__get_param_file_path_from_key(context, key) is None

###--------------------------------------------------------------------------------------------

    def __assert_param_file_exists(self, context, key):
        if context:
            if not self.__context_exists(context):
                # raise DSOException(CLI_MESSAGES['ContextNotFound'].format(context))
                raise DSOException(CLI_MESSAGES['ParameterNotFound'].format(key))

        if not self.__param_group_exists(context, key):
            # raise DSOException("Parameter group '{0}' not found.".format(self.__get_param_group_from_key(key)))
            raise DSOException(CLI_MESSAGES['ParameterNotFound'].format(key))

###--------------------------------------------------------------------------------------------

    def __load_param_file_from_key(self, context, key):
        paramFilePath = self.__get_param_file_path_from_key(context, key)
        if paramFilePath:
            return self.__load_param_file(paramFilePath)
        else:
            return None

###--------------------------------------------------------------------------------------------

    ### get inherited filepath
    def __locate_param_file_from_key(self, context, key):
        ### first check if overide param file exists in context
        path = self.__get_param_file_path_from_key(context, key)
        if not path and context:
            stage, env = Stages.parse_context(context)
            ### if not found, check default env in the stage
            if not env == 'defaut':
                path = self.__get_param_file_path_from_key(f"{stage}/default", key)
                ### if not found, check param root path
                if not path:
                    path = self.__get_param_file_path_from_key('', key)
        return path

###--------------------------------------------------------------------------------------------
###--------------------------------------------------------------------------------------------

    def get_key_validator(self):

        allowGroups = Configs.parameter_spec('allowGroups')
        if allowGroups is None:
            Logger.debug(f"'allowGroups' is not set for the parameter provider, defaulted to '{defaults['allowGroups']}'.")
            allowGroups = defaults['allowGroups']

        if allowGroups:
            return r"^([a-zA-Z][a-zA-Z0-9]*/)?([a-zA-Z][a-zA-Z0-9_.-]*)$"
        else:
            return r"^([a-zA-Z][a-zA-Z0-9_.-]*)$"




###--------------------------------------------------------------------------------------------

    def add(self, context, key, value):
        paramFilePath = self.__get_param_file_path_from_key(context, key, create=True)
        params = self.__load_param_file(paramFilePath)
        paramName = self.__get_param_name_from_key(key)
        set_dict_value(params, paramName.split('.'), value)
        with open(paramFilePath, 'w') as configFile:
            yaml.dump(params, configFile, sort_keys=False, indent=2)

###--------------------------------------------------------------------------------------------

    def list(self, context, uninherited):

        # if not self.__context_exists(context):
        #     if uninherited:
        #         Logger.warn(CLI_MESSAGES['ContextNotFound'].format(context))
        #     else:
        #         Logger.warn(CLI_MESSAGES['ContextNotFoundListingInherited'].format(context))

        grouppedParameters = {}

        if uninherited:
            ### iterating through param files in root or stage/env and merge with paramDict
            path = self.__get_params_root_path(context)
            for fileName in self.__list_param_file_names(path):
                Logger.debug("Loading '{0}/{1}'".format(path, fileName))
                group = self.__parse_group_from_file_name(fileName)
                params = self.__load_param_file(os.path.join(path, fileName))
                grouppedParameters[group] = params

        ### if not uninherited:
        else:
            path = self.__get_params_root_path()
            ### iterating through parameters root files and add them to paramDict
            for fileName in self.__list_param_file_names(path):
                Logger.debug("Loading '{0}/{1}'".format(path, fileName))
                group = self.__parse_group_from_file_name(fileName)
                params = self.__load_param_file(os.path.join(path, fileName))
                if group in grouppedParameters.keys():
                    grouppedParameters[group] = merge_dicts(params, grouppedParameters[group])
                else:
                    grouppedParameters[group] = params

            if context: 
                stage, env = Stages.parse_context(context)
                
                ### iterating through param files in stage(/default) and add them to paramDict
                path = self.__get_params_root_path(f"{stage}/default")
                for fileName in self.__list_param_file_names(path):
                    Logger.debug("Loading '{0}/{1}'".format(path, fileName))
                    group = self.__parse_group_from_file_name(fileName)
                    params = self.__load_param_file(os.path.join(path, fileName))
                    if group in grouppedParameters.keys():
                        grouppedParameters[group] = merge_dicts(params, grouppedParameters[group])
                    else:
                        grouppedParameters[group] = params

                if not env == 'default':
                    ### iterating through param files in stage/env and merge them paramDict
                    path = self.__get_params_root_path(f"{stage}/{env}")
                    for fileName in self.__list_param_file_names(path):
                        Logger.debug("Loading '{0}/{1}'...".format(path, fileName))
                        group = self.__parse_group_from_file_name(fileName)
                        params = self.__load_param_file(os.path.join(path, fileName))
                        if group in grouppedParameters.keys():
                            grouppedParameters[group] = merge_dicts(params, grouppedParameters[group])
                        else:
                            grouppedParameters[group] = params

        prependGroups = Configs.parameter_spec('prependGroups')
        if prependGroups is None:
            Logger.debug(f"'prependGroups' is not set for the parameter provider, defaulted to '{defaults['prependGroups']}'.")
            prependGroups = defaults['prependGroups']

        groupDelimiter = Configs.parameter_spec('groupDelimiter') 
        if groupDelimiter is None:
            Logger.debug(f"'groupDelimiter' is not set for the parameter provider, defaulted to '{defaults['groupDelimiter']}'.")
            groupDelimiter = defaults['groupDelimiter']

        nestedDelimiter = Configs.parameter_spec('nestedDelimiter')
        if nestedDelimiter is None:
            Logger.debug(f"'nestedDelimiter' is not set for the parameter provider, defaulted to '{defaults['nestedDelimiter']}'.")
            nestedDelimiter = defaults['nestedDelimiter']

        # for group, params in grouppedParameters.items():
        #     for key, value in params.items():
        #         grouppedkey = f"{group}{groupDelimiter}{key}" if prependGroups and not group == 'default_group' else key
        #         if grouppedkey in parameters.keys():
        #             Logger.warn(f"'{group}/{key}' was ignored as it would duplicate '{grouppedkey}'.")
        #         set_dict_value(parameters, grouppedkey.split('.'), value)


        # if not nestedDelimiter == '.':
        #     parameters = flatten_dict(parameters, delimiter=nestedDelimiter)

        ### first flatten params, if only nestedDelimiter is not '.', otherwise jinja2 will ot be able to resolve dic nested items
        if not nestedDelimiter == '.':
            for group, params in grouppedParameters.items():
                grouppedParameters[group] = flatten_dict(params, delimiter=nestedDelimiter)

        parameters = {}
        for group, params in grouppedParameters.items():
            for key, value in params.items():
                grouppedkey = f"{group}{groupDelimiter}{key}" if prependGroups and not group == 'default_group' else key
                if grouppedkey in parameters.keys():
                    Logger.warn(f"'{group}/{key}' was ignored as it would duplicate '{grouppedkey}'.")
                set_dict_value(parameters, grouppedkey.split('.'), value)

        return parameters


###--------------------------------------------------------------------------------------------

    def get(self, context, key):
        path = self.__locate_param_file_from_key(context, key)
        if not path:
                raise DSOException(CLI_MESSAGES['ParameterNotFound'].format(key))
        params = self.__load_param_file(path)
        value = get_dict_item(params, self.__get_param_name_from_key(key).split('.'))
        if value is None or isinstance(value, dict):
            raise DSOException(CLI_MESSAGES['ParameterNotFound'].format(key))
        # if isinstance(value, dict):
        #     raise DSOException(CLI_MESSAGES['ParameterNotFoundScope'].format(key))
        return value

###--------------------------------------------------------------------------------------------

    def delete(self, context, key):
        self.__assert_param_file_exists(context, key)
        params = self.__load_param_file_from_key(context, key)
        keys = self.__get_param_name_from_key(key).split('.')
        if not del_dict_item(params, keys):
            raise DSOException(CLI_MESSAGES['ParameterNotFound'].format(key))
        ### recursively delet empty parent scopes 
        del_dict_empty_item(params, keys[:-1])
        path = self.__get_param_file_path_from_key(context, key)
        if len(params.keys()) > 0:
            self.__save_param_file(params, path)
        else:
            os.remove(path)
            conetxt_root_path = self.__get_params_root_path(context)
            if len(os.listdir(conetxt_root_path)) == 0:
                os.rmdir(conetxt_root_path)


